/* Dazuko FreeBSD. Allow FreeBSD file access control for 3rd-party applications.
   Written by John Ogness <jogness@antivir.de>

   Copyright (c) 2003 H+BEDV Datentechnik GmbH
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. Neither the name of Dazuko nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/


#include <sys/types.h>
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/proc.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/syscall.h>
#include <sys/filio.h>
#include <machine/stdarg.h>

#include <sys/namei.h>
#include <sys/param.h>
#include <sys/vnode.h>

#include "dazuko_freebsd.h"
#include "dazuko_xp.h"
#include "dazukoio.h"


d_read_t freebsd_dazuko_device_read;
d_write_t freebsd_dazuko_device_write;
d_ioctl_t freebsd_dazuko_device_ioctl;
d_open_t freebsd_dazuko_device_open;
d_close_t freebsd_dazuko_device_close;

extern struct xp_atomic active;

static int	dev_major = 33;
static dev_t	sdev = NULL;

#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_open(struct proc *p, struct open_args *uap);
static struct sysent freebsd_dazuko_sys_open_sysent = { 3, (sy_call_t*)freebsd_dazuko_sys_open };
static sy_call_t *original_sys_open;
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_close(struct proc *p, struct close_args *uap);
static struct sysent freebsd_dazuko_sys_close_sysent = { 1, (sy_call_t*)freebsd_dazuko_sys_close };
static sy_call_t *original_sys_close;
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
static int freebsd_dazuko_sys_write(struct proc *p, struct write_args *uap);
static struct sysent freebsd_dazuko_sys_write_sysent = { 3, (sy_call_t*)freebsd_dazuko_sys_write };
static sy_call_t *original_sys_write;
#endif

#ifdef ON_EXEC_SUPPORT
static int freebsd_dazuko_sys_execve(struct proc *p, struct execve_args *uap);
static struct sysent freebsd_dazuko_sys_execve_sysent = { 3, (sy_call_t*)freebsd_dazuko_sys_execve };
static sy_call_t *original_sys_execve;
#endif

#ifdef ON_UNLINK_SUPPORT
static int freebsd_dazuko_sys_unlink(struct proc *p, struct unlink_args *uap);
static struct sysent freebsd_dazuko_sys_unlink_sysent = { 1, (sy_call_t*)freebsd_dazuko_sys_unlink };
static sy_call_t *original_sys_unlink;
#endif

#ifdef ON_RMDIR_SUPPORT
static int freebsd_dazuko_sys_rmdir(struct proc *p, struct rmdir_args *uap);
static struct sysent freebsd_dazuko_sys_rmdir_sysent = { 1, (sy_call_t*)freebsd_dazuko_sys_rmdir };
static sy_call_t *original_sys_rmdir;
#endif

static struct vnode	*orig_rootmnt = NULL;

static struct cdevsw	cdepts = {
					freebsd_dazuko_device_open,
					freebsd_dazuko_device_close,
					freebsd_dazuko_device_read,
					freebsd_dazuko_device_write,
					freebsd_dazuko_device_ioctl,
					nopoll,
					nommap,
					nostrategy,
					DEVICE_NAME,
					33,
					nodump,
					nopsize,
					D_TTY,
					-1
				};


MALLOC_DECLARE(M_DAZUKOBUF);
MALLOC_DEFINE(M_DAZUKOBUF, "dazukobuffer", "buffer for dazuko module");


static inline int freebsd_getpid(struct proc *p)
{
	if (p == NULL)
	{
		xp_print("dazuko: warning: freebsd_getpid(NULL)\n");
		return -1;
	}

	return p->p_pid;
}


/* mutex */

inline int xp_init_mutex(struct xp_mutex *mutex)
{
	if (mutex == NULL)
	{
		xp_print("dazuko: warning: xp_init_mutex(NULL)\n");
		return -1;
	}

	simple_lock_init(&(mutex->lock));

	return 0;
}

inline int xp_down(struct xp_mutex *mutex)
{
	if (mutex == NULL)
	{
		xp_print("dazuko: warning: xp_down(NULL)\n");
		return -1;
	}

	simple_lock(&(mutex->lock));

	return 0;
}

inline int xp_up(struct xp_mutex *mutex)
{
	if (mutex == NULL)
	{
		xp_print("dazuko: warning: xp_up(NULL)\n");
		return -1;
	}

	simple_unlock(&(mutex->lock));

	return 0;
}


/* read-write lock */

inline int xp_init_rwlock(struct xp_rwlock *rwlock)
{
	if (rwlock == NULL)
	{
		xp_print("dazuko: warning: xp_init_rwlock(NULL)\n");
		return -1;
	}

	lockinit(&(rwlock->lock), PVM, "dazuko", 0, 0);

	return 0;
}

inline int xp_write_lock(struct xp_rwlock *rwlock)
{
	if (rwlock == NULL)
	{
		xp_print("dazuko: warning: xp_write_lock(NULL)\n");
		return -1;
	}

	lockmgr(&(rwlock->lock), LK_EXCLUSIVE, 0, curproc);

	return 0;
}

inline int xp_write_unlock(struct xp_rwlock *rwlock)
{
	if (rwlock == NULL)
	{
		xp_print("dazuko: warning: xp_write_unlock(NULL)\n");
		return -1;
	}

	lockmgr(&(rwlock->lock), LK_RELEASE, 0, curproc);

	return 0;
}

inline int xp_read_lock(struct xp_rwlock *rlock)
{
	if (rlock == NULL)
	{
		xp_print("dazuko: warning: xp_read_lock(NULL)\n");
		return -1;
	}

	lockmgr(&(rlock->lock), LK_SHARED, 0, curproc);

	return 0;
}

inline int xp_read_unlock(struct xp_rwlock *rlock)
{
	if (rlock == NULL)
	{
		xp_print("dazuko: warning: xp_read_unlock(NULL)\n");
		return -1;
	}

	lockmgr(&(rlock->lock), LK_RELEASE, 0, curproc);

	return 0;
}


/* wait-notify queue */

inline int xp_init_queue(struct xp_queue *queue)
{
	return 0;
}

inline int xp_wait_until_condition(struct xp_queue *queue, int (*cfunction)(void *), void *cparam, int allow_interrupt)
{
	int	rc = 0;
	int	prio = PVM;

	if (allow_interrupt)
		prio |= PCATCH;

	for (;;)
	{
		if (asleep(queue, prio, "dazuko", 0) != 0)
		{
			rc = -1;
			break;
		}

		if (cfunction(cparam))
			break;

		if (await(prio, 0) != 0)
		{
			rc = -1;
			break;
		}
	}

	asleep(NULL, 0, "", 0);

	return rc;
}

inline int xp_notify(struct xp_queue *queue)
{
	wakeup(queue);

	return 0;
}


/* memory */

inline void* xp_malloc(int size)
{
	void	*ret;

	ret = malloc(size, M_DAZUKOBUF, M_WAITOK);

	if (ret == NULL)
		xp_print("dazuko: warning: malloc(%d, ...) -> NULL\n", size);

	return ret;
}

inline int xp_free(void *ptr)
{
	if (ptr == NULL)
	{
		xp_print("dazuko: warning: xp_free(NULL)\n");
		return 0;
	}

	free(ptr, M_DAZUKOBUF);

	return 0;
}


/* path attribute */

inline int xp_is_absolute_path(const char *path)
{
	if (path == NULL)
	{
		xp_print("dazuko: warning: xp_is_absolute_path(NULL)\n");
		return 0;
	}

	return (path[0] == '/');
}


/* atomic */

inline int xp_atomic_set(struct xp_atomic *atomic, int value)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_set(NULL, %d)\n", value);
		return -1;
	}

	atomic->value = value;

	return 0;
}

inline int xp_atomic_inc(struct xp_atomic *atomic)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_inc(NULL)\n");
		return -1;
	}

	(atomic->value)++;

	return 0;
}

inline int xp_atomic_dec(struct xp_atomic *atomic)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_dec(NULL)\n");
		return -1;
	}

	(atomic->value)--;

	return 0;
}

inline int xp_atomic_read(struct xp_atomic *atomic)
{
	if (atomic == NULL)
	{
		xp_print("dazuko: error: xp_atomic_read(NULL)\n");
		return -1;
	}

	return atomic->value;
}


/* file descriptor */

inline int xp_copy_file(struct xp_file *dest, struct xp_file *src)
{
	if (dest == NULL || src == NULL)
	{
		xp_print("dazuko: error: xp_copy_file(%s, %s)\n", dest ? "dest" : "NULL", src ? "src" : "NULL");
		return -1;
	}

	dest->file = src->file;

	return 0;
}

inline int xp_compare_file(struct xp_file *file1, struct xp_file *file2)
{
	if (file1 == NULL || file2 == NULL)
	{
		xp_print("dazuko: error: xp_compare_file(%s, %s)\n", file1 ? "file1" : "NULL", file2 ? "file2" : "file2");
		return -1;
	}

	return (file1->file != file2->file);
}


/* file structure */

static void freebsd_get_full_fileinfo(struct proc *p, const char *path, char **fullpath, char **freefullpath, struct file_properties *file_p)
{
	struct proc			localp;
	struct nameidata		nd;
	struct filedesc			localfd;
	struct vattr			vattr;
/*
int temp;
char kpath[256];
*/

	if (p == NULL || path == NULL || fullpath == NULL || freefullpath == NULL || file_p == NULL)
		return;

	dazuko_bzero(&nd, sizeof(nd));
	dazuko_bzero(&localp, sizeof(localp));
	dazuko_bzero(&localfd, sizeof(localfd));
	dazuko_bzero(file_p, sizeof(struct file_properties));

	NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_USERSPACE, path, p);

	if (namei(&nd) != 0)
	{
		/*
		we need to understand why this fails sometimes
		for (temp=0 ; fubyte(path+temp) ; temp++);
		if (temp > 255) temp = 255;
		copyin(path, kpath, temp);
		kpath[temp] = 0;
		xp_print("dazuko: error: namei failed on %s! (%d)\n", kpath, temp);
		*/
	}
	else
	{
		localp.p_textvp = nd.ni_vp;
		localp.p_fd = &localfd;
		if (orig_rootmnt)
			localfd.fd_rdir = orig_rootmnt;
		else
			localfd.fd_rdir = nd.ni_rootdir;

		textvp_fullpath(&localp, fullpath, freefullpath);

		if (nd.ni_vp != NULL)
		{
			if (VOP_GETATTR(nd.ni_vp, &vattr, p->p_ucred, p) == 0)
			{
				file_p->size = vattr.va_size;
				file_p->set_size = 1;
				file_p->uid = vattr.va_uid;
				file_p->set_uid = 1;
				file_p->gid = vattr.va_gid;
				file_p->set_gid = 1;
				file_p->mode = vattr.va_mode;
				file_p->set_mode = 1;
				file_p->device_type = vattr.va_fsid;
				file_p->set_device_type = 1;
			}

			/* deref looked up vnode */
			vrele(nd.ni_vp);
		}
	}
}

inline int xp_file_struct_check(struct dazuko_file_struct *dfs)
{
	char	*fullpath = NULL;
	char	*freefullpath = NULL;
	int	length;

	if (dfs == NULL)
	{
		xp_print("dazuko: warning: xp_file_struct_check(NULL)\n");
		return -1;
	}

	freebsd_get_full_fileinfo(dfs->extra_data->p, dfs->extra_data->user_filename, &fullpath, &freefullpath, &(dfs->file_p));

	if (freefullpath)
	{
		length = dazuko_get_filename_length(fullpath);

		dfs->filename = (char *)xp_malloc(length + 1);
		if (dfs->filename)
		{
			memcpy(dfs->filename, fullpath, length);
			dfs->filename[length] = 0;
			dfs->filename_length = length;  /* the string length */
		}

/*
		xp_free(freefullpath);
*/
		free(freefullpath, M_TEMP);
	}

	if (dfs->filename == NULL)
	{
		for (length=0 ; fubyte((dfs->extra_data->user_filename)+length) ; length++);

		length++;

		dfs->filename = (char *)xp_malloc(length);
		if (dfs->filename == NULL)
			return -1;

		if (copyin(dfs->extra_data->user_filename, dfs->filename, length) != 0)
		{
			xp_print("dazuko: error: xp_file_struct_check.copyin() failed\n");
			xp_free(dfs->filename);
			return -1;
		}

		dfs->filename_length = length - 1;  /* the string length */
	}

	return 0;
}

inline int xp_file_struct_check_cleanup(struct dazuko_file_struct *dfs)
{
	return 0;
}

static int dazuko_file_struct_cleanup(struct dazuko_file_struct **dfs)
{
	if (!dfs)
		return 0;

	if (!*dfs)
		return 0;

	xp_file_struct_check_cleanup(*dfs);

	if ((*dfs)->filename)
		xp_free((*dfs)->filename);

	if ((*dfs)->extra_data)
		xp_free((*dfs)->extra_data);

	xp_free(*dfs);

	*dfs = NULL;

	return 0;
}


/* system calls */

#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_open(struct proc *p, struct open_args *uap)
{
	struct dazuko_file_struct	*dfs = NULL;
	struct event_properties		event_p;
	int				error = 0;
	int				check_error = 0;
#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	struct xp_file			file;
#endif

	if (p == NULL)
	{
		xp_print("dazuko: warning: freebsd_dazuko_sys_open(NULL, ...)\n");
		check_error = -1;
	}
	else
	{
		check_error = dazuko_sys_check(DAZUKO_ON_OPEN, 1, freebsd_getpid(p));
	}

	if (!check_error)
	{
		dazuko_bzero(&event_p, sizeof(event_p));

		if (uap == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_open(p, NULL)\n");
			check_error = -1;
		}
		else if (p->p_cred == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_open.p->p_cred=NULL\n");
			check_error = -1;
		}
		else
		{
			event_p.flags = uap->flags;
			event_p.set_flags = 1;
			event_p.mode = uap->mode;
			event_p.set_mode = 1;
			event_p.pid = freebsd_getpid(p);
			event_p.set_pid = 1;
			event_p.uid = p->p_cred->p_ruid;
			event_p.set_uid = 1;

			dfs = (struct dazuko_file_struct *)xp_malloc(sizeof(struct dazuko_file_struct));
			if (dfs)
			{
				dazuko_bzero(dfs, sizeof(struct dazuko_file_struct));

				dfs->extra_data = (struct xp_file_struct *)xp_malloc(sizeof(struct xp_file_struct));
				if (dfs->extra_data)
				{
					dazuko_bzero(dfs->extra_data, sizeof(struct xp_file_struct));

					dfs->extra_data->user_filename = uap->path;
					dfs->extra_data->p = p;

					error = dazuko_sys_pre(DAZUKO_ON_OPEN, dfs, &event_p);
				}
				else
				{
					xp_free(dfs);
					dfs = NULL;
				}
			}
		}
	}

	if (error)
	{
		/* access should be blocked */

		error = EPERM;
	}
	else
	{
		/* call the standard open function */
		error = original_sys_open(p, uap);

		#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
		{
			if (!check_error)
			{
				if (error)
				{
					check_error = -1;
				}
				else if (p->p_fd == NULL)
				{
					check_error = -1;
				}
				else if (p->p_retval[0] < 0)
				{
					check_error = -1;
				}
				else if (p->p_retval[0] >= p->p_fd->fd_nfiles)
				{
					check_error = -1;
				}
				else if (p->p_fd->fd_ofiles == NULL)
				{
					check_error = -1;
				}
				else if (p->p_fd->fd_ofiles[p->p_retval[0]] == NULL)
				{
					check_error = -1;
				}
				else
				{
					file.file = p->p_fd->fd_ofiles[p->p_retval[0]];
				}
			}
		}
		#endif
	}

	if (!check_error)
	{
		#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
		{
			if (!error && dfs)
				dazuko_sys_post(DAZUKO_ON_OPEN, dfs, &file, &event_p);
		}
		#endif
	}

	dazuko_file_struct_cleanup(&dfs);

	return error;
}
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
static int freebsd_dazuko_sys_close(struct proc *p, struct close_args *uap)
{
	/* The kernel wants to close the given file
	 * descriptor. */

	int			error = 0;
	int			check_error = 0;
	struct xp_file		file;
	struct event_properties	event_p;

	if (p == NULL)
	{
		xp_print("dazuko: warning: freebsd_dazuko_sys_close(NULL, ...)\n");
		check_error = -1;
	}
	else
	{
		check_error = dazuko_sys_check(DAZUKO_ON_CLOSE, 1, freebsd_getpid(p));
	}

	if (!check_error)
	{
		if (uap == NULL)
		{
			check_error = -1;
		}
		else if (p->p_fd == NULL)
		{
			check_error = -1;
		}
		else if (uap->fd < 0)
		{
			check_error = -1;
		}
		else if ((u_int)(uap->fd) >= p->p_fd->fd_nfiles)
		{
			check_error = -1;
		}
		else if (p->p_fd->fd_ofiles == NULL)
		{
			check_error = -1;
		}
		else if (p->p_fd->fd_ofiles[uap->fd] == NULL)
		{
			check_error = -1;
		}
		else
		{
			/* grab the file* for possible later use */
			file.file = p->p_fd->fd_ofiles[uap->fd];
		}
	}

	error = original_sys_close(p, uap);

	if (!error && !check_error)
	{
		if (p->p_cred == NULL)
		{
			xp_print("dazuko: error: freebsd_dazuko_sys_close.p->p_cred=NULL\n");
		}
		else
		{
			dazuko_bzero(&event_p, sizeof(event_p));
			event_p.pid = freebsd_getpid(p);
			event_p.set_pid = 1;
			event_p.uid = p->p_cred->p_ruid;
			event_p.set_uid = 1;

			dazuko_sys_post(DAZUKO_ON_CLOSE, NULL, &file, &event_p);
		}
	}

	return error;
}
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
static int freebsd_dazuko_sys_write(struct proc *p, struct write_args *uap)
{
	/* The kernel wants to write to the given file
	 * descriptor. */

	int			error = 0;
	int			check_error = 0;
	struct xp_file		file;
	struct event_properties	event_p;

	check_error = dazuko_sys_check(DAZUKO_ON_CLOSE_MODIFIED, 1, freebsd_getpid(p));

	if (!check_error)
	{
		if (uap == NULL)
		{
			check_error = -1;
		}
		else if (p->p_fd == NULL)
		{
			check_error = -1;
		}
		else if (uap->fd < 0)
		{
			check_error = -1;
		}
		else if ((u_int)(uap->fd) >= p->p_fd->fd_nfiles)
		{
			check_error = -1;
		}
		else if (p->p_fd->fd_ofiles == NULL)
		{
			check_error = -1;
		}
		else if (p->p_fd->fd_ofiles[uap->fd] == NULL)
		{
			check_error = -1;
		}
		else
		{
			/* grab the file* for possible later use */
			file.file = p->p_fd->fd_ofiles[uap->fd];
		}
	}

	error = original_sys_write(p, uap);
	
	/* did we actually write anything? */

	if (!error && !check_error)
	{
		dazuko_bzero(&event_p, sizeof(event_p));

		dazuko_sys_post(DAZUKO_ON_CLOSE_MODIFIED, NULL, &file, &event_p);
	}

	return error;
}
#endif

static inline int freebsd_dazuko_sys_generic(int event, struct proc *p, const char *user_filename, int daemon_is_allowed)
{
	struct dazuko_file_struct	*dfs = NULL;
	struct event_properties		event_p;
	int				error = 0;
	int				check_error = 0;

	check_error = dazuko_sys_check(event, daemon_is_allowed, freebsd_getpid(p));

	if (check_error == 0)
	{
		if (p == NULL || user_filename == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_generic(%d, %s, %s)\n", event, p ? "p" : "NULL", user_filename ? "user_filename" : "NULL");
			check_error = -1;
		}
		else if (p->p_cred == NULL)
		{
			xp_print("dazuko: warning: freebsd_dazuko_sys_generic.p->p_cred=NULL\n");
			check_error = -1;
		}
		else
		{
			dazuko_bzero(&event_p, sizeof(event_p));
			event_p.pid = freebsd_getpid(p);
			event_p.set_pid = 1;
			event_p.uid = p->p_cred->p_ruid;
			event_p.set_uid = 1;

			dfs = (struct dazuko_file_struct *)xp_malloc(sizeof(struct dazuko_file_struct));
			if (dfs)
			{
				dazuko_bzero(dfs, sizeof(struct dazuko_file_struct));

				dfs->extra_data = (struct xp_file_struct *)xp_malloc(sizeof(struct xp_file_struct));
				if (dfs->extra_data)
				{
					dazuko_bzero(dfs->extra_data, sizeof(struct xp_file_struct));

					dfs->extra_data->user_filename = user_filename;

					error = dazuko_sys_pre(event, dfs, &event_p);

					dazuko_file_struct_cleanup(&dfs);
				}
				else
				{
					xp_free(dfs);
					dfs = NULL;
				}
			}
		}
	}

	return error;
}

#ifdef ON_EXEC_SUPPORT
static int freebsd_dazuko_sys_execve(struct proc *p, struct execve_args *uap)
{
	int	error;

	error = freebsd_dazuko_sys_generic(DAZUKO_ON_EXEC, p, uap->fname, 0);

	if (error)
		return error;

	return original_sys_execve(p, uap);
}
#endif

#ifdef ON_UNLINK_SUPPORT
static int freebsd_dazuko_sys_unlink(struct proc *p, struct unlink_args *uap)
{
	int	error;

	error = freebsd_dazuko_sys_generic(DAZUKO_ON_UNLINK, p, uap->path, 1);

	if (error)
		return error;

	return original_sys_unlink(p, uap);
}
#endif

#ifdef ON_RMDIR_SUPPORT
static int freebsd_dazuko_sys_rmdir(struct proc *p, struct rmdir_args *uap)
{
	int	error;

	error = freebsd_dazuko_sys_generic(DAZUKO_ON_RMDIR, p, uap->path, 1);

	if (error)
		return error;

	return original_sys_rmdir(p, uap);
}
#endif


/* system hook */

inline int xp_sys_hook()
{
	sdev = make_dev(&cdepts, 0, UID_ROOT, GID_WHEEL, 0600, DEVICE_NAME);

#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	DPRINT(("dazuko: hooked sys_open\n"));
	original_sys_open = sysent[SYS_open].sy_call;
	sysent[SYS_open] = freebsd_dazuko_sys_open_sysent;
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
        DPRINT(("dazuko: hooked sys_close\n"));
	original_sys_close = sysent[SYS_close].sy_call;
	sysent[SYS_close] = freebsd_dazuko_sys_close_sysent;
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
        DPRINT(("dazuko: hooked sys_write\n"));
	original_sys_write = sysent[SYS_write].sy_call;
	sysent[SYS_write] = freebsd_dazuko_sys_write_sysent;
#endif

#ifdef ON_EXEC_SUPPORT
        DPRINT(("dazuko: hooked sys_execve\n"));
	original_sys_execve = sysent[SYS_execve].sy_call;
	sysent[SYS_execve] = freebsd_dazuko_sys_execve_sysent;
#endif

#ifdef ON_UNLINK_SUPPORT
        DPRINT(("dazuko: hooked sys_unlink\n"));
	original_sys_unlink = sysent[SYS_unlink].sy_call;
	sysent[SYS_unlink] = freebsd_dazuko_sys_unlink_sysent;
#endif

#ifdef ON_RMDIR_SUPPORT
        DPRINT(("dazuko: hooked sys_rmdir\n"));
	original_sys_rmdir = sysent[SYS_rmdir].sy_call;
	sysent[SYS_rmdir] = freebsd_dazuko_sys_rmdir_sysent;
#endif

	return 0;
}

inline int xp_sys_unhook()
{
#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	if (sysent[SYS_open].sy_call != freebsd_dazuko_sys_open_sysent.sy_call)
		xp_print("dazuko: open system call not correct (system may be left in an unstable state!)\n");
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	if (sysent[SYS_close].sy_call != freebsd_dazuko_sys_close_sysent.sy_call)
		xp_print("dazuko: close system call not correct (system may be left in an unstable state!)\n");
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
	if (sysent[SYS_write].sy_call != freebsd_dazuko_sys_write_sysent.sy_call)
		xp_print("dazuko: write system call not correct (system may be left in an unstable state!)\n");
#endif

#ifdef ON_EXEC_SUPPORT
	if (sysent[SYS_execve].sy_call != freebsd_dazuko_sys_execve_sysent.sy_call)
		xp_print("dazuko: execve system call not correct (system may be left in an unstable state!)\n");
#endif

#ifdef ON_UNLINK_SUPPORT
	if (sysent[SYS_unlink].sy_call != freebsd_dazuko_sys_unlink_sysent.sy_call)
		xp_print("dazuko: unlink system call not correct (system may be left in an unstable state!)\n");
#endif

#ifdef ON_RMDIR_SUPPORT
	if (sysent[SYS_rmdir].sy_call != freebsd_dazuko_sys_rmdir_sysent.sy_call)
		xp_print("dazuko: rmdir system call not correct (system may be left in an unstable state!)\n");
#endif


#if defined(ON_OPEN_SUPPORT) || defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	sysent[SYS_open].sy_call = (sy_call_t*)original_sys_open;
#endif

#if defined(ON_CLOSE_SUPPORT) || defined(ON_CLOSE_MODIFIED_SUPPORT)
	sysent[SYS_close].sy_call = (sy_call_t*)original_sys_close;
#endif

#ifdef ON_CLOSE_MODIFIED_SUPPORT
	sysent[SYS_write].sy_call = (sy_call_t*)original_sys_write;
#endif

#ifdef ON_EXEC_SUPPORT
	sysent[SYS_execve].sy_call = (sy_call_t*)original_sys_execve;
#endif

#ifdef ON_UNLINK_SUPPORT
	sysent[SYS_unlink].sy_call = (sy_call_t*)original_sys_unlink;
#endif

#ifdef ON_RMDIR_SUPPORT
	sysent[SYS_rmdir].sy_call = (sy_call_t*)original_sys_rmdir;
#endif

	destroy_dev(sdev);

	return 0;
}


/* output */

inline int xp_print(const char *fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);

	return 0;
}


/* device file operations */

int freebsd_dazuko_device_open(dev_t dev, int oflags, int devtype, struct proc *p)
{
	DPRINT(("dazuko: freebsd_dazuko_device_open() [%d]\n", freebsd_getpid(p)));

	return 0;
}

static int read_dev_major(struct uio *uio)
{
	size_t	dev_major_len;
	char	tmp[20];

	if (dev_major < 0)
		return ENODEV;

	if (uio->uio_offset != 0)
		return 0;

	/* print dev_major to a string
 	* and get length */
	dazuko_bzero(tmp, sizeof(tmp));

	dev_major_len = snprintf(tmp, sizeof(tmp), "%d", dev_major);

	if (tmp[sizeof(tmp)-1] != 0)
	{
		xp_print("dazuko: failing device_read, device number overflow for dameon %d (dev_major=%d)\n", freebsd_getpid(uio->uio_procp), dev_major);
		return EFAULT;
	}

	if (uio->uio_iov == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read.uio->uio_iov=NULL\n");
		return EFAULT;
	}

	if (uio->uio_iov->iov_len < dev_major_len)
		return EINVAL;

	/* copy dev_major string to userspace */
	if (uiomove(tmp, dev_major_len, uio) != 0)
		return EFAULT;

	return 0;
}

int freebsd_dazuko_device_read(dev_t dev, struct uio *uio, int ioflag)
{
	if (uio == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read(..., NULL, ...)\n");
		return EFAULT;
	}

	if (uio->uio_procp == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read.uio->uio_procp=NULL\n");
		return EFAULT;
	}

	if (uio->uio_procp->p_cred == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_read.uio->uio_procp->p_cred=NULL\n");
		return EFAULT;
	}

	DPRINT(("dazuko: freebsd_dazuko_device_read() [%d]\n", freebsd_getpid(uio->uio_procp)));

	/* non-root daemons are ignored */
	if (uio->uio_procp->p_cred->p_ruid != 0)
		return 0;

	/* return dev_major if process is not registered */
	if (!dazuko_is_our_daemon(freebsd_getpid(uio->uio_procp)))
		return read_dev_major(uio);

	return 0;
}

static inline int protocol13(struct dazuko_request *u_request, struct proc *p)
{
	int			error = 0;
	struct dazuko_request	*request;

	/* allocate kernel request */
	request = (struct dazuko_request *)xp_malloc(sizeof(struct dazuko_request));
	if (request == NULL)
		return EFAULT;

/* use out1 now */

	/* copy in the request */
	if (copyin(u_request->type, request->type, sizeof(char[2])) != 0)
	{
		error = EFAULT;
		goto freebsd_dazuko_device_ioctl_out1;
	}

	/* copy buffer size from userspace */
	if (copyin(&(u_request->buffer_size), &(request->buffer_size), sizeof(int)) != 0)
	{
		error = EFAULT;
		goto freebsd_dazuko_device_ioctl_out1;
	}

	/* sanity check */
	if (request->buffer_size < 0 || request->buffer_size > 8192)
	{
		error = EFAULT;
		goto freebsd_dazuko_device_ioctl_out1;
	}

	/* copy reply buffer size from userspace */
	if (copyin(&(u_request->reply_buffer_size), &(request->reply_buffer_size), sizeof(int)) != 0)
	{
		error = EFAULT;
		goto freebsd_dazuko_device_ioctl_out1;
	}

	/* sanity check */
	if (request->reply_buffer_size < 0 || request->reply_buffer_size > 8192)
	{
		error = EPERM;
		goto freebsd_dazuko_device_ioctl_out1;
	}

	/* allocate buffer */
	request->buffer = (char *)xp_malloc(request->buffer_size + 1);
	if (request->buffer == NULL)
	{
		error = EFAULT;
		goto freebsd_dazuko_device_ioctl_out1;
	}

/* use out2 now */

	if (request->reply_buffer_size > 0)
	{
		/* allocate reply buffer */
		request->reply_buffer = (char *)xp_malloc(request->reply_buffer_size + 1);
		if (request->reply_buffer == NULL)
		{
			error = EFAULT;
			goto freebsd_dazuko_device_ioctl_out2;
		}

/* use out3 now */

		request->reply_buffer_size_used = 0;
	}

	/* copy the buffer from userspace to kernelspace */
	if (copyin(u_request->buffer, request->buffer, request->buffer_size) != 0)
	{
		error = EFAULT;
		goto freebsd_dazuko_device_ioctl_out3;
	}

	request->buffer[request->buffer_size] = 0;

	error = dazuko_handle_request(freebsd_getpid(p), request);

	if (error == 0 && request->reply_buffer_size > 0)
	{
		request->reply_buffer[request->reply_buffer_size] = 0;

		if (copyout(&(request->reply_buffer_size_used), &(u_request->reply_buffer_size_used), sizeof(int)) != 0)
		{
			error = EFAULT;
			goto freebsd_dazuko_device_ioctl_out3;
		}

		if (request->reply_buffer_size_used > 0)
		{
			if (copyout(request->reply_buffer, u_request->reply_buffer, request->reply_buffer_size_used) != 0)
			{
				error = EFAULT;
				goto freebsd_dazuko_device_ioctl_out3;
			}
		}
	}

freebsd_dazuko_device_ioctl_out3:
	if (request->reply_buffer_size > 0)
		xp_free(request->reply_buffer);
freebsd_dazuko_device_ioctl_out2:
	xp_free(request->buffer);
freebsd_dazuko_device_ioctl_out1:
	xp_free(request);

	return error;
}

int freebsd_dazuko_device_write(dev_t dev, struct uio *uio, int ioflag)
{
	struct dazuko_request	*u_request;
	char			buffer[32];
	char			*value;
	int			size;

	if (uio == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_write(..., NULL, ...)\n");
		return EFAULT;
	}

	if (uio->uio_procp == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_write.uio->uio_procp=NULL\n");
		return EFAULT;
	}

	if (uio->uio_procp->p_cred == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_write.uio->uio_procp->p_cred=NULL\n");
		return EFAULT;
	}

	/* non-root daemons are ignored */
	if (uio->uio_procp->p_cred->p_ruid != 0)
		return uio->uio_resid;

	size = uio->uio_resid;
	if (size >= sizeof(buffer))
		size = sizeof(buffer) - 1;

	/* copy request_pointer string to kernelspace */
	if (uiomove(buffer, size, uio) != 0)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_write.uiomove!=0\n");
		return EFAULT;
	}

	buffer[size] = 0;

	if (dazuko_get_value("\nRA=", buffer, &value) != 0)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_write.RA missing\n");
		return EFAULT;
	}

	u_request = (struct dazuko_request *)strtoul(value, NULL, 10);

	xp_free(value);

	if (protocol13(u_request, uio->uio_procp) == 0)
		return uio->uio_resid;
	else
		return EINTR;
}

static int freebsd_dazuko_device_ioctl_compat12(dev_t dev, u_long cmd, caddr_t data, int fflag, struct proc *p)
{
	int			error = 0;
	int			len;
	int			opt;
	struct access_compat12	*u_dazuko;
	struct option_compat12	*u_option;
	struct slot		*s;
	char			*k_param;
	int			i;
	struct daemon_id		did;

	did.pid = freebsd_getpid(p);
	did.unique = -1;

	/* A macro is used to translate the cmd argument.
	 * This keeps it compatible across various Linux
	 * platforms. */
	switch (cmd)
	{
		case IOCTL_GET_AN_ACCESS: /* deprecated */
			/* The daemon is requesting a filename of a file
			 * to scan. This code will wait until a filename
			 * is available, or until we should be killed.
			 * (killing is done if any errors occur as well
			 * as when the user kills us) */

			u_dazuko = *(struct access_compat12 **)data;

/* DOWN? */
			s = dazuko_get_an_access(&did);

			if (s == NULL)
				return EINTR;

/* DOWN */

			/* Slot IS in DAZUKO_WORKING state. Copy all the
			 * necessary information to userspace structure. */

			if (copyout(s->filename, u_dazuko->filename, (s->filenamelength)+1) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->event), &(u_dazuko->event), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->event_p.flags), &(u_dazuko->o_flags), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->event_p.mode), &(u_dazuko->o_mode), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->event_p.uid), &(u_dazuko->uid), sizeof(int)) != 0)
			{
				error = EFAULT;
			}
			else if (copyout(&(s->event_p.pid), &(u_dazuko->pid), sizeof(int)) != 0)
			{
				error = EFAULT;
			}

			xp_up(&(s->mutex));
/* UP */

			if (error)
			{
				dazuko_state_error(s, DAZUKO_WORKING);
			}

			return error;

		case IOCTL_RETURN_ACCESS: /* deprecated */
			/* The daemon has finished scanning a file
			 * and has the response to give. The daemon's
			 * slot should be in the DAZUKO_WORKING state. */

			u_dazuko = *(struct access_compat12 **)data;

			/* copy the response into the slot */
			if (copyin(&(u_dazuko->deny), &i, sizeof(int)) != 0)
			{
				return EFAULT;
			}

			return dazuko_return_access(&did, i, NULL);

		case IOCTL_SET_OPTION: /* deprecated */
			/* The daemon wants to set a configuration
			 * option in the kernel. */

			u_option = *(struct option_compat12 **)data;

			/* copy option type from userspace */
			if (copyin(&(u_option->command), &opt, sizeof(int)) != 0)
				return EPERM;

			/* copy path length from userspace */
			if (copyin(&(u_option->buffer_length), &len, sizeof(int)) != 0)
				return EPERM;

			/* sanity check */
			if (len < 0 || len > 4096)
				return EPERM;

			k_param = (char *)xp_malloc(len + 1);
			if (!k_param)
				return EFAULT;

			/* We must copy the param from userspace to kernelspace. */

			if (copyin(u_option->buffer, k_param, len) != 0)
			{
				xp_free(k_param);
				return EFAULT;
			}

			k_param[len] = 0;

			if (opt == REGISTER)
				error = dazuko_register_daemon(&did, k_param, len, 1);
			else
				error = dazuko_set_option(&did, opt, k_param, len);

			xp_free(k_param);

			return error;

		default:
			xp_print("dazuko: daemon %d requested unknown device_ioctl %c,%d (possible bug)\n", freebsd_getpid(p), (cmd >> 8) & 0xff, cmd & 0xff);

			break;
	}

	return 0;
}

int freebsd_dazuko_device_ioctl(dev_t dev, u_long cmd, caddr_t data, int fflag, struct proc *p)
{
	/* A daemon uses this function to interact with
	 * the kernel. A daemon can set scanning parameters,
	 * give scanning response, and get filenames to scan. */

	int	error = 0;

	if (p == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_ioctl(..., NULL)\n");
		return EFAULT;
	}

	if (p->p_cred == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_ioctl.p->p_cred=NULL\n");
		return EFAULT;
	}

	/* non-root daemons are ignored */
	if (p->p_cred->p_ruid != 0)
		return 0;

	if (data == NULL)
	{
		xp_print("dazuko: error: freebsd_dazuko_device_ioctl(..., NULL, ...)\n");
		return EFAULT;
	}

	switch (cmd)
	{
		case IOCTL_REQUEST:
			/* Protocol 13 */

			error = protocol13(*(struct dazuko_request **)data, p);

			break;

		case FIONBIO:
			error = 0;
			break;

		case FIOASYNC:
			error = 0;
			break;

		default:
			/* compat12 */

			error = freebsd_dazuko_device_ioctl_compat12(dev, cmd, data, fflag, p);

			break;
	}

	if (error < 0)
	{
		/* general error occurred */

		return EPERM;
	}

	return error;
}

int freebsd_dazuko_device_close(dev_t dev, int fflag, int devtype, struct proc *p)
{
	struct daemon_id	did;

	DPRINT(("dazuko: dazuko_device_close() [%d]\n", freebsd_getpid(p)));

	/* non-root daemons are ignored */
	if (p->p_cred->p_ruid != 0)
		return 0;

	did.pid = freebsd_getpid(p);
	did.unique = -1;

	/* note: This only works for compat12 mode.
	 * For 1.3, the daemon must properly unregister.
	 */

	return dazuko_unregister_daemon(&did);
}


/* init/exit */

static int dazuko_loader(struct module *m, int what, void *arg)
{
	int			err = 0;
	struct nameidata	nd;

	switch (what)
	{
		case MOD_LOAD:
			err = dazuko_init();

			if (!err)
			{
				NDINIT(&nd, LOOKUP, NOFOLLOW, UIO_SYSSPACE, "/", curproc);

				if (namei(&nd))
					xp_print("dazuko: warning: failed to get root mount\n");
				else
					orig_rootmnt = nd.ni_rootdir;

				/* deref looked up vnode */
				if (nd.ni_vp)
					vrele(nd.ni_vp);
			}
			break;

		case MOD_UNLOAD:
			if (dazuko_exit() != 0)
				err = EPERM;
			break;

		default:
			err = EINVAL;
			break;
	}

	return err;
}

DEV_MODULE(dazuko, dazuko_loader, NULL);

