/* Example program demonstrating the multi-thread capabilities/interface
   of Dazuko.
   Written by John Ogness <jogness@antivir.de>

   Copyright (c) 2003 H+BEDV Datentechnik GmbH
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

   3. Neither the name of Dazuko nor the names of its contributors may be used
   to endorse or promote products derived from this software without specific
   prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

/* This program is a simple application demonstrating how to interface
   with Dazuko usint the multi-thread interface (_TS). The program
   instructs Dazuko to detect all types of accesses within the specified
   directories (or any subdirectory thereof). The program than prints out
   the accesses and instructs Dazuko to allow access.

   This program will run only after Dazuko has been successfully installed.
   Please see the Dazuko website for information on how to do this.
   http://www.dazuko.org

   Note: For security reasons, Dazuko will only interact with applications
         running as root. Therefore, this example program must also be run
         as root.
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include "dazukoio.h"

struct main_loop_struct
{
	int	argc;
	char	**argv;
	int	thread_id;
};

#define NUMBER_OF_THREADS 5

int			RUNNING = 1;		/* flag for main loop */
struct main_loop_struct	id[NUMBER_OF_THREADS]; 	/* id's for threads */

void print_usage(void)
{
	printf("usage: example_mt <dir> <dir> ...\n");
}

void sigterm(int sig)
{
	/* exit the loop on TERM or INT signal */

	printf("detected signal: %d\n", sig);

	RUNNING = 0;
	signal(sig, sigterm);
}

void print_access(struct dazuko_access *acc, int thread_id)
{
	/* print access data */

	switch (acc->event)
	{
		case DAZUKO_ON_OPEN:
			printf("OPEN ");
			break;
		case DAZUKO_ON_CLOSE:
			printf("CLOSE ");
			break;
		case DAZUKO_ON_CLOSE_MODIFIED:
			printf("CLOSE (modified) ");
			break;
		case DAZUKO_ON_EXEC:
			printf("EXEC ");
			break;
		default:
			printf("???? ");
			break;
	}
	printf("flags:%d ", acc->flags);
	printf("mode:%d ", acc->mode);
	printf("uid:%d ", acc->uid);
	printf("pid:%d ", acc->pid);
	printf("file:%s ", acc->filename);
	printf(" (thread %d)", thread_id);
	printf("\n");

	fflush(stdout);
}

void* main_loop(void *param)
{
	dazuko_id_t		*dazuko_id;
	struct dazuko_access	*acc;
	int			thread_id = 0;
	int			args_ok = 0;
	int			i;
	int			argc;
	char			**argv;

	if (param == NULL)
	{
		printf("error: invalid thread parameter\n");
		return NULL;
	}

	argc = ((struct main_loop_struct *)param)->argc;
	argv = ((struct main_loop_struct *)param)->argv;
	thread_id = ((struct main_loop_struct *)param)->thread_id;

	/* register with dazuko */
	if (dazukoRegister_TS(&dazuko_id, "DAZUKO_MT_EXAMPLE", "r+") != 0)
	{
		printf("error: failed to register with Dazuko (thread %d)\n", thread_id);
		return NULL;
	}

	printf("registered with Dazuko successfully (thread %d)\n", thread_id);

	/* set access mask */
	if (dazukoSetAccessMask_TS(dazuko_id, DAZUKO_ON_OPEN | DAZUKO_ON_CLOSE | DAZUKO_ON_CLOSE_MODIFIED | DAZUKO_ON_EXEC) != 0)
	{
		printf("error: failed to set access mask (thread %d)\n", thread_id);
		dazukoUnregister_TS(&dazuko_id);
		return NULL;
	}

	printf("set access mask successfully (thread %d)\n", thread_id);

	/* set scan path */
	for (i=1 ; i<argc ; i++)
	{
		if (argv[i][0] == '/')
		{
			if (dazukoAddIncludePath(argv[i]) != 0)
			{
				printf("error: failed to add %s include path (thread %d)\n", argv[i], thread_id);
				dazukoUnregister_TS(&dazuko_id);
				return NULL;
			}

			args_ok = 1;
		}
	}

	if (!args_ok)
	{
		print_usage();
		dazukoUnregister_TS(&dazuko_id);
		return NULL;
	}

	printf("set scan path successfully (thread %d)\n", thread_id);

	while (RUNNING)
	{
		/* get an access */
		if (dazukoGetAccess_TS(dazuko_id, &acc) == 0)
		{
			print_access(acc, thread_id);

			/* always allow access */
			acc->deny = 0;

			/* return access (IMPORTANT, the kernel is waiting for us!) */
			if (dazukoReturnAccess_TS(dazuko_id, &acc) != 0)
			{
				printf("error: failed to return access (thread %d)\n", thread_id);
				RUNNING = 0;
			}
		}
		else
		{
			printf("warning: failed to get an access (thread %d)\n", thread_id);
			RUNNING = 0;
		}
	}

	/* unregister with Dazuko */
	if (dazukoUnregister_TS(&dazuko_id) != 0)
	{
		printf("error: failed to unregister with Dazuko (thread %d)\n", thread_id);
		return NULL;
	}

	printf("unregistered with Dazuko successfully (thread %d)\n", thread_id);
 
	return NULL;
}

int main(int argc, char *argv[])
{
	pthread_t	tid[NUMBER_OF_THREADS];
	int		i;

	/* make sure we are root */
	if (getuid() != 0)
	{
		printf("only root can run this program\n");
		return -1;
	}

	/* detect TERM signals */
	signal(SIGTERM, sigterm);

	/* detect INT signals */
	signal(SIGINT, sigterm);

	for (i=0 ; i<NUMBER_OF_THREADS ; i++)
	{
		/* set the id for the thread */
		id[i].thread_id = i;

		/* set argc,argv for the thread */
		id[i].argc = argc;
		id[i].argv = argv;

		/* create the thread */
		pthread_create(&(tid[i]), NULL, main_loop, &(id[i]));
	}

	for (i=0 ; i<NUMBER_OF_THREADS ; i++)
	{
		/* wait for the thread to finish */
		pthread_join(tid[i], NULL);
	}

	return 0;
}
